#include "gaamscomm.h"

#define CMD_READER_CLEAR 0
#define CMD_READER_READ 1
#define CMD_READER_GREEN 2
#define CMD_READER_RED 3
#define CMD_READER_OPEN 16

// MODRL12
// p=0 - ziskaj stav
// p=1 - nastav stav natrvalo
// p>=2 - nastav na cas 0.5, 1, 2, 3, 4 sekundy
//                        pppsssss0sssssss000mmmmm0mmmmmmm
#define CMD_MODRL_CLEAR  0b00100000000000000001111101111111
#define CMD_MODRL_READ   0b00000000000000000000000000000000
#define CMD_MODRL_STATE  0b01000000000000000000000000000000

#define CMD_MID_CLEAR 0b01100000
#define CMD_MID_SET   0b01110000
#define CMD_MID_READ  0b01000000

#define MID_SET_BYTE 0x05

//#define REP_MASK_DISC 128
#define REP_MASK_RED 64
#define REP_MASK_GREEN 32
#define REP_MASK_OPEN 16
#define REP_MASK_CARD 8
#define REP_MASK_CMD 23
#define REP_MASK_ERROR 256

volatile int progend = 0;
volatile int gaamscomm_clear_counter = 0;
sem_t gaamscomm_answer_semaphore;
pthread_t gaamscomm_answer_thr[NUM_ANSWER_THREADS];
pthread_mutex_t gaamscomm_mutex = PTHREAD_MUTEX_INITIALIZER;
INFO info;

void p_printf (const char *fmt, ...)
{
	char buf[4096], fname[PATH_MAX];
	struct tm *t;
	time_t theTime;
	FILE *s;
	va_list ap;

	va_start (ap, fmt);
	vsnprintf (buf, sizeof (buf), fmt, ap);
	va_end (ap);
	buf[sizeof (buf) - 1] = '\0';

	pthread_mutex_lock (&gaamscomm_mutex);

	time (&theTime);
	t = localtime (&theTime);

	if (info.logpath[0] != '\0') {
		snprintf (fname, sizeof (fname), "%s/%04d-%02d-%02d.txt", info.logpath, t->tm_year + 1900, t->tm_mon + 1, t->tm_mday);
		s = fopen (fname, "a");
		if (s) {
			fprintf (s, "%04d-%02d-%02d %02d:%02d:%02d %s\n", t->tm_year + 1900, t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec, buf);
			fclose (s);
		}
	}
	if (info.localecho) {
		printf ("%s\n", buf);
	}

	pthread_mutex_unlock (&gaamscomm_mutex);
}

void gaamscomm_destination_begin (void)
{
	/* First, we need to split info.url into parts. */
	info.url_proto = NULL;
	info.url_host = NULL;
	info.url_port = NULL;
	info.url_path = NULL;

	int cnt = 0, i, l, pathl = 0;
	char *tempurl = strdup (info.url);
	if (tempurl == NULL) {
		main_exit ("Malloc failed!");
	}

	char *tok = strtok (tempurl, "/");
	while (tok) {
		if (cnt == 0) {
			info.url_proto = strdup (tok);
			if (info.url_proto == NULL) {
				main_exit ("Malloc failed!");
			}
		}
		else if (cnt == 1) {
			info.url_host = strdup (tok);
			if (info.url_host == NULL) {
				main_exit ("Malloc failed!");
			}

			l = strlen (tok);
			for (i = 0; i < l - 1; i++) {
				if (tok[i] == ':') {
					info.url_host[i] = '\0';

					info.url_port = strdup (tok + i + 1);
					if (info.url_port == NULL) {
						main_exit ("Malloc failed!");
					}

					break;
				}
			}
		}

		if (cnt >= 1) {
			l = strlen (tok);
			info.url_path = realloc (info.url_path, pathl + 2 + l);
			if (info.url_path == NULL) {
				main_exit ("Realloc failed!");
			}

			info.url_path[pathl] = '/'; pathl++;
			memcpy (info.url_path + pathl, tok, l); pathl += l;
			info.url_path[pathl] = '\0';
		}

		tok = strtok (NULL, "/");
		cnt ++;
	}
	free (tempurl);

//	printf ("url: '%s'\n", info.url);
//	printf ("host: '%s'\n", info.url_host);
//	printf ("port: '%s'\n", info.url_port);
//	printf ("proto: '%s'\n", info.url_proto);
//	printf ("path: '%s'\n", info.url_path);

	info.use_what = USE_FILE;
	int ret;

	if (strcasecmp (info.url_proto, "http:") == 0) {
		if (http_global_open (0) != 0) {
			main_exit ("Could not initialize libcurl!");
		}
		info.use_what = USE_HTTP;
	}
	else if (strcasecmp (info.url_proto, "https:") == 0) {
		if (http_global_open (1) != 0) {
			main_exit ("Could not initialize libcurl with SSL support!");
		}
		info.use_what = USE_HTTP;
	}
	else if (strcasecmp (info.url_proto, "system:") == 0) {
		info.use_what = USE_SYSTEM;

	}
	else if (strcasecmp (info.url_proto, "pipe:") == 0) {
		info.use_what = USE_PIPE;

#ifndef OS_WIN
		umask (0111);
		if (mkfifo (info.url_path, 0666) == -1) {
			if (errno != EEXIST) {
				main_exit ("Could not initialize named pipe '%s' for writing", info.url_path);
			}
		}
#endif
	}
	else if (strcasecmp (info.url_proto, "tcp:") == 0) {
		ret = tcpclient_global_open ();
		if (ret != 0) {
			main_exit ("Could not initialize TCP socket! Retcode: %d", ret);
		}
		info.use_what = USE_TCPCLIENT;

	}
	else if (strcasecmp (info.url_proto, "listen:") == 0) {
		ret = tcpserver_global_open ();
		if (ret != 0) {
			main_exit ("Could not initialize TCP listening socket! Retcode: %d", ret);
		}
		info.use_what = USE_TCPSERV;
	}

}

void gaamscomm_destination_end (void)
{
	if (info.use_what == USE_HTTP) {
		http_global_close();
	}
	if (info.use_what == USE_TCPCLIENT) {
		tcpclient_global_close();
	}
	if (info.use_what == USE_TCPSERV) {
		tcpserver_global_close();
	}
}

int8_t gaamscomm_prepare_output (const GAAMSCOMM_EVENT *ev, char *buf, const size_t sz)
{
	int i, l;

	buf[0] = '\0';
	if (info.type_node[ev->ccc] == TYPE_READER) {
		if (ev->num_card < 1) return -5;

		snprintf (buf, sz - 1,
				"gaamscomm_type=reader"
				"&entry_reader=%02X"
				"&entry_hwid=%s"
				"&entry_type=%c",
				info.id_node[ev->ccc], ev->card[0], ev->card[0][0]);

		for (i = 0; i < (int) ev->num_card;  i++) {
			l = strlen (buf);
			int w = snprintf (buf + l, sz - l, "&entries_hwid[]=%s", ev->card[i]);
			if (w >= (int) (sz - l)) {
				return -3;
			}
			buf[l + w] = '\0';
		}
	} else if (info.type_node[ev->ccc] == TYPE_MODRL) {
		snprintf (buf, sz - 1,
				"gaamscomm_type=modrl"
				"&entry_modrl=%02X"
				"&entry_state=",
				info.id_node[ev->ccc]);

		l = strlen (buf);
		for (i = 0; i < 13; i++) {
			if (i == 7) continue;
			buf[l] = ( (ev->bitmap >> i) & 1) ? 'c' : 'o';
			l++;
		}
		buf[l] = '\0';
	} else if (info.type_node[ev->ccc] == TYPE_MID) {
		snprintf (buf, sz - 1,
				"gaamscomm_type=mid"
				"&entry_mid=%02X"
				"&entry_port=%d"
				"&entry_state=%d"
				"&entry_counter=%d",
				info.id_node[ev->ccc], ( (ev->bitmap >> 24) & 0xff) + 1, ev->bitmap & 1, ( ( (ev->bitmap >> 16) & 0x7f) << 13) | ( ( (ev->bitmap >> 8) & 0x7f) << 6) | ( (ev->bitmap >> 1) & 0x3f));
	} else if (info.type_node[ev->ccc] == TYPE_SYSTEM) {
		snprintf (buf, sz - 1,
				"gaamscomm_type=system"
				 );
		if (ev->bitmap == SYSTEM_EVENT_CLEAR) {
			strcat (buf, "&entry_event=clear");
		}
	} else {
		return -6;
	}

	if (info.entry_id[0] != '\0') {
		l = strlen (buf);
		sprintf (buf + l, "&entry_id=%s", info.entry_id);
	}

	return 0;
}

int8_t gaamscomm_prepare_tcp_output (const GAAMSCOMM_EVENT *ev, char *buf, const size_t sz)
{
	int i, l;

	buf[0] = '\0';
	if (info.type_node[ev->ccc] == TYPE_READER) {
		if (ev->num_card < 1) return -2;

		for (i = 0; i < (int) ev->num_card;  i++) {
			l = strlen (buf);
			int w = snprintf (buf + l, sz - l, "%s\r\n", ev->card[i]);
			if (w >= (int) (sz - l)) {
				return -3;
			}
			buf[l + w] = '\0';
		}

		return 0;
	}

	return -1;
}

uint8_t gaamscomm_reverse_byte (const uint8_t in)
{
	uint8_t out = 0;

	int i;
	for (i = 0; i < 8; i++) {
		out <<= 1;
		out |= (in >> i) & 1;
	}

	return out;
}

void gaamscomm_reverse_buffer (const uint8_t *inbuf, uint8_t *outbuf, const size_t len, const int really)
{
	if (!really) {
		memcpy (outbuf, inbuf, len);
		return;
	}

	size_t i;
	for (i = 0; i < len; i++) {
		outbuf[i]= gaamscomm_reverse_byte (inbuf[len - (i + 1)]);
	}
}

void gaamscomm_format_card_id (uint8_t *outbuf, CARD_TYPE type, const uint8_t *inbuf, const size_t len)
{
	if (type == CARD_OLD) {
		if (info.card_output == OUTPUT_GAAMS) {
			sprintf ( (char *) outbuf, "G%02X%02X%02X%02X%02X", inbuf[0], inbuf[1], inbuf[2], inbuf[3], inbuf[4]);
		} if (info.card_output == OUTPUT_STRIPPED || info.card_output == OUTPUT_REVERSED) {
			uint8_t buf[5];
			gaamscomm_reverse_buffer (inbuf, buf, 5, info.card_output == OUTPUT_REVERSED);
			sprintf ( (char *) outbuf, "%02X%02X%02X%02X%02X", buf[0], buf[1], buf[2], buf[3], buf[4]);
		}
	} else if (type == CARD_MIFARE) {
		if (info.card_output == OUTPUT_GAAMS) {
			outbuf[0] = 'M';
			outbuf[1] = '\0';
			char tmp[8];
			size_t i;
			for (i = 0; i < len; i++) {
				sprintf (tmp, "%02X", inbuf[i]);
				strcat ( (char *) outbuf, tmp);
			}
		} if (info.card_output == OUTPUT_STRIPPED || info.card_output == OUTPUT_REVERSED) {
			uint8_t buf[256];
			char tmp[8];
			size_t i;
			gaamscomm_reverse_buffer (inbuf + 1, buf, len - 1, info.card_output == OUTPUT_REVERSED);
			for (i = 0; i < len - 1; i++) {
				sprintf (tmp, "%02X", buf[i]);
				strcat ( (char *) outbuf, tmp);
			}
		}
	}
}

int8_t gaamscomm_send (const uint8_t ccc, const uint32_t cmd, const size_t sz, const uint8_t *inbuf, const size_t insz)
{
	uint8_t buf[BUFSIZE];
	ssize_t l = 0;
	int i;

	buf[l] = BYTE_STX;
	l++;
	buf[l] = info.id_node[ccc] & 0xff;
	l++;
	if (sz != 1 && sz != 4) return -1;

	for (i = sz - 1; i >= 0; i--) {
		buf[l] = (cmd >> (i * 8)) & 255;
		l++;
	}

	if (insz > 0) {
		if (inbuf == NULL) return -1;
		if (insz + l + 1 >= BUFSIZE) return -1;
		for (i = 0; i < (int) insz; i++) {
			buf[l] = inbuf[i];
			l++;
		}
	}

	buf[l] = 0x58;
	for (i = 0; i < (int) l; i++) {
		buf[l] ^= buf[i];
	}
	l++;

	if (rs485_send (l, buf) != l) {
		info.writeErrorCount++;
		if (info.writeErrorCount > MAXwriteErrorCount) {
			progend = PROGEND_writeErrorCount;
		}
		return -2;
	}
	info.writeErrorCount = 0;

	return 0;
}

int8_t gaamscomm_recv (const uint8_t ccc, uint8_t *outbuf)
{
	uint8_t buf[BUFSIZE], crc = 0x58;
	ssize_t i, l, now = 0;

	l = rs485_recv (buf);
	if (l <= 0) {
		info.noResponseCount++;
		if (info.noResponseCount > MAXnoResponseCount) {
			progend = PROGEND_noResponseCount;
		}
		return -1;
	}
	info.noResponseCount = 0;

	if (l >= 2 && buf[0] == BYTE_STX) {
		if (buf[1] != (info.id_node[ccc] & 0xff)) {
			if (info.noResponseCount > MAXnoResponseCount) {
				progend = PROGEND_noResponseCount;
			}
			p_printf ("BAD DATA");
			return -2;
		}
		now = 2;
	}

	for (i = 0; i < l - 1; i++) {
		crc ^= buf[i];
	}
	if (crc != buf[l - 1]) {
		info.readErrorCount++;
		if (info.readErrorCount > MAXreadErrorCount) {
			progend = PROGEND_readErrorCount;
		}
		return -4;
	}
	l--;

	if (info.type_node[ccc] == TYPE_READER) {

		if (l != (1 + now) && l < (6 + now)) {
			info.readErrorCount++;
			if (info.readErrorCount > MAXreadErrorCount) {
				progend = PROGEND_readErrorCount;
			}
			return -3;
		}

		if (outbuf != NULL) {
			outbuf[0] = '\0';
			if (buf[now]&REP_MASK_CARD) {
				if (l == 6 + now) {
					gaamscomm_format_card_id (outbuf, CARD_OLD, buf + (now + 1), 5);
				} else if (buf[now + 1] == 'M' && l > 6 + now) {
					gaamscomm_format_card_id (outbuf, CARD_MIFARE, buf + (now + 2), l - (now + 2));
				} else {
					info.readErrorCount++;
					if (info.readErrorCount > MAXreadErrorCount) {
						progend = PROGEND_readErrorCount;
					}
					return -5;
				}
			}
		}

		info.readErrorCount = 0;
		return buf[now];
	}

	if (info.type_node[ccc] == TYPE_MODRL) {

		if (l != now + 2) {
			info.readErrorCount++;
			if (info.readErrorCount > MAXreadErrorCount) {
				progend = PROGEND_readErrorCount;
			}
			return -3;
		}

		if (outbuf != NULL) {
			outbuf[0] = buf[now];
			outbuf[1] = buf[now + 1];
		}

		info.readErrorCount = 0;
		return 0;
	}

	if (info.type_node[ccc] == TYPE_MID) {

		if (l != now + 49) {
			info.readErrorCount++;
			if (info.readErrorCount > MAXreadErrorCount) {
				progend = PROGEND_readErrorCount;
			}
			return -3;
		}

		if (outbuf != NULL) {
			memcpy (outbuf, buf + now + 1, 48);
		}

		info.readErrorCount = 0;
		return buf[now];
	}

	// Ak sme sa dostali sem, tak sme nerozoznali typ zariadenia
	info.readErrorCount++;
	if (info.readErrorCount > MAXreadErrorCount) {
		progend = PROGEND_readErrorCount;
	}

	return -99;
}

int32_t gaamscomm_reader (const uint8_t ccc, const uint8_t cmd)
{
	int32_t ret;
	char card[128];
	GAAMSCOMM_EVENT ev;
	queue_new_event (&ev);
	ev.ccc = ccc;

	while (1) {
		if (gaamscomm_send (ccc, cmd, 1, NULL, 0) != 0) {
			ret = -10;
			break;
		}
		ret = gaamscomm_recv (ccc, (uint8_t *) card);
		if (info.debug) p_printf ("gaamscomm_reader: gaamscomm_recv returned %d", ret);
		if (ret < 0) {
			break;
		}
		if ( (ret & REP_MASK_CMD) != cmd) {
			if (info.debug) p_printf ("gaamscomm_reader: command mismatch: %d != %d", (ret & REP_MASK_CMD), cmd);
			ret = -11;
			break;
		}

		if (ret & REP_MASK_CARD) {
			if (ev.num_card < CARDBUF_SIZE) {
				ev.card[ev.num_card] = strdup (card);
				if (ev.card[ev.num_card] == NULL) {
					ret = -12;
					break;
				}
				p_printf ("Card #%d: %s", ev.num_card + 1, card);
				ev.num_card++;
			}
		} else break;
	}
	if (ret < 0) {
		queue_destroy_event (&ev);
		return ret;
	}
	if (ev.num_card > 0) {
		if (info.debug) p_printf ("Sending READER event to thread from %02X with %d cards", info.id_node[ev.ccc], ev.num_card);
		queue_push (&info.q_in, &ev);
		sem_post (&gaamscomm_answer_semaphore);
	}

	return ret;
}

int32_t gaamscomm_modrl (const uint8_t ccc, const uint32_t cmd)
{
	int32_t ret;
	uint8_t buf[2];
	GAAMSCOMM_EVENT ev;

	if (gaamscomm_send (ccc, cmd, 4, NULL, 0) != 0) {
		return -10;
	}
	ret = gaamscomm_recv (ccc, buf);
	if (info.debug) p_printf ("gaamscomm_modrl: gaamscomm_recv returned %d", ret);
	if (ret < 0) {
		return ret;
	}
	if ( (cmd >> 29) != (buf[0] >> 5)) {
		return -11;
	}

	if (cmd == CMD_MODRL_STATE) {
		queue_new_event (&ev);
		ev.ccc = ccc;
		ev.bitmap = ( (buf[0] & 0x1f) << 8) | buf[1];
		if (info.debug) p_printf ("Sending MODRL response to thread from %02X with %u bitmap", info.id_node[ev.ccc], ev.bitmap);
		queue_push (&info.q_in, &ev);
		sem_post (&gaamscomm_answer_semaphore);
	}

	return 0;
}

int32_t gaamscomm_mid (const uint8_t ccc, const uint32_t cmd)
{
	int32_t ret, i;
	uint8_t buf[48];
	GAAMSCOMM_EVENT ev[16];

	if ( (cmd & 0xff) == CMD_MID_SET) {
		for (i = 0; i < 16; i++) {
			buf[i] = (cmd >> 8) & 0xff;
		}
		if (gaamscomm_send (ccc, cmd & 0xff, 1, buf, 16) != 0) {
			return -10;
		}
	} else {
		if (gaamscomm_send (ccc, cmd & 0xff, 1, NULL, 0) != 0) {
			return -10;
		}
	}

	ret = gaamscomm_recv (ccc, buf);
	if (info.debug) p_printf ("gaamscomm_mid: gaamscomm_recv returned %d", ret);
	if (ret < 0) {
		return ret;
	}

	if (info.buf_node[ccc] == NULL) return -11;
	if (!memcmp (info.buf_node[ccc], buf, 48)) return 0;
	for (i = 0; i < 16; i++) {
		if (memcmp (info.buf_node[ccc] + (i * 3), buf + (i * 3), 3)) {
			queue_new_event (&ev[i]);
			ev[i].ccc = ccc;
			ev[i].bitmap = (i << 24) | (buf[i * 3] << 16) | (buf[i * 3 + 1] << 8) | buf[i * 3 + 2];
			if (info.debug) p_printf ("Sending MID event to thread from %02X with %u bitmap", info.id_node[ev[i].ccc], ev[i].bitmap);
			queue_push (&info.q_in, &ev[i]);
			sem_post (&gaamscomm_answer_semaphore);
		}
	}
	memcpy (info.buf_node[ccc], buf, 48);

//        c=((int)buf[i*3]<<13)|((int)buf[i*3+1]<<6)|((int)buf[i*3+2]>>1);
//        p_printf("MID %02X, vstup %02d: poc=%d, s=%d",info.id_node[ccc],i+1,c,buf[i*3+2]&1);

	return 0;
}

int gaamscomm_clear (GAAMSCOMM_LOOP_STATE *ls)
{
	int is_present = 0;
	GAAMSCOMM_EVENT ev;
	int32_t id_system = -1;

	gaamscomm_clear_counter++;

	info.noResponseCount = 0;
	info.readErrorCount = 0;
	info.writeErrorCount = 0;

	memset (info.enabled, 0, sizeof (info.enabled));

	// let's clean all buffers
	for (ls->ccc = 0; ls->ccc < info.num_nodes; ls->ccc++) {
		if (info.id_node[ls->ccc] >= READER_START && info.id_node[ls->ccc] < READER_END) {
			info.type_node[ls->ccc] = TYPE_READER;
			info.buf_node[ls->ccc] = NULL;
			ls->ret = gaamscomm_reader (ls->ccc, CMD_READER_CLEAR);
			if (ls->ret >= 0) {
				p_printf ("Buffer of reader %02X cleared successfully", info.id_node[ls->ccc]);
				if (ls->ret >= 0) {
					info.enabled[ls->ccc] = 1;
					p_printf ("Reader %02X is PRESENT", info.id_node[ls->ccc]);
					is_present = 1;
				} else {
					info.enabled[ls->ccc] = 0;
					p_printf ("Reader %02X is not connected", info.id_node[ls->ccc]);
				}
			} else {
				p_printf ("Buffer of reader %02X clear FAILED!!!", info.id_node[ls->ccc]);
			}
		}
		if (info.id_node[ls->ccc] >= MODRL_START && info.id_node[ls->ccc] < MODRL_END) {
			info.type_node[ls->ccc] = TYPE_MODRL;
			if (info.buf_node[ls->ccc] == NULL) info.buf_node[ls->ccc] = (uint8_t *) malloc (2);
			memset (info.buf_node[ls->ccc], 0, 2);
			ls->ret = gaamscomm_modrl (ls->ccc, CMD_MODRL_CLEAR);
			if (ls->ret >= 0) {
				p_printf ("Buffer of modrl %02X cleared successfully", info.id_node[ls->ccc]);
				if (ls->ret >= 0) {
					info.enabled[ls->ccc] = 1;
					p_printf ("Modrl %02X is PRESENT", info.id_node[ls->ccc]);
					is_present = 1;
				} else {
					info.enabled[ls->ccc] = 0;
					p_printf ("Modrl %02X is not connected", info.id_node[ls->ccc]);
				}
			} else {
				p_printf ("Buffer of modrl %02X clear FAILED!!!", info.id_node[ls->ccc]);
			}
		}
		if (info.id_node[ls->ccc] >= MID_START && info.id_node[ls->ccc] < MID_END) {
			info.type_node[ls->ccc] = TYPE_MID;
			if (info.buf_node[ls->ccc] == NULL) info.buf_node[ls->ccc] = (uint8_t *) malloc (48);
			memset (info.buf_node[ls->ccc], 0, 48);
			ls->ret = gaamscomm_mid (ls->ccc, CMD_MID_CLEAR);
			if (ls->ret >= 0) {
				p_printf ("Buffer of mid %02X cleared successfully", info.id_node[ls->ccc]);
				if (ls->ret >= 0) {
					info.enabled[ls->ccc] = 1;
					p_printf ("Mid %02X is PRESENT", info.id_node[ls->ccc]);
					is_present = 1;
				} else {
					info.enabled[ls->ccc] = 0;
					p_printf ("Mid %02X is not connected", info.id_node[ls->ccc]);
				}
			} else {
				p_printf ("Buffer of mid %02X clear FAILED!!!", info.id_node[ls->ccc]);
			}
		}
		if (info.id_node[ls->ccc] >= SYSTEM_START && info.id_node[ls->ccc] < SYSTEM_END) {
			info.type_node[ls->ccc] = TYPE_SYSTEM;
			info.buf_node[ls->ccc] = NULL;
			info.enabled[ls->ccc] = 1;
			id_system = ls->ccc;
		}
	}

	ls->ccc = 0;
	ls->look_counter = 0;

	queue_clear (&info.q_in);
	queue_clear (&info.q_out);

	gaamscomm_clear_counter++;

	if (id_system >= 0) {
		queue_new_event (&ev);
		ev.ccc = id_system;
		ev.bitmap = SYSTEM_EVENT_CLEAR;
		queue_push (&info.q_in, &ev);
		sem_post (&gaamscomm_answer_semaphore);
	}

	return is_present;
}

int gaamscomm_one_loop (GAAMSCOMM_LOOP_STATE *ls)
{
	GAAMSCOMM_EVENT ev;
	int is_present = 0;
	uint8_t ccc = 0;
	uint32_t mask;

	while (!progend) {
		mask = 0;

		if (queue_front (&info.q_out, &ev)) {
			if (abs (ev.timestamp - time (0)) < MAX_TIME_ELAPSED) {
				ccc = ev.ccc;
				mask = ev.bitmap;
				if (info.debug) p_printf ("LOOP: Received event for %02X with bitmap %X", info.id_node[ev.ccc], ev.bitmap);
			} else {
				p_printf ("LOOP: Received event for %02X, but it's TOO OLD!", info.id_node[ev.ccc]);
			}
			queue_destroy_event (&ev);
		}

		if (mask == 0) {
			if (ls->ccc >= info.num_nodes) {
				ls->look_counter++;

				if (ls->look_counter > RECHECK_AFTER_LOOPS) {
					ls->look_counter = 0;

					for (ls->ccc = 0; ls->ccc < info.num_nodes; ls->ccc++) {
						if (!info.enabled[ls->ccc]) {
							info.enabled[ls->ccc] = 2;
						}
					}
				}

				ls->ccc = 0;
				break;
			}
			ccc = ls->ccc;
			ls->ccc++;
		}

		// info.enabled[ccc]==2 znamena, ze ma zistit, ci je online alebo nie
		if (info.enabled[ccc] == 2 || info.enabled[ccc] == 1 /*&& (info.type_node[ccc]==TYPE_READER||info.type_node[ccc]==TYPE_READER) ) */) {

			if (info.type_node[ccc] == TYPE_READER) {
				if (info.enabled[ccc] == 1) {
					ls->ret = gaamscomm_reader (ccc, mask ? (mask & 255) : CMD_READER_READ);
				} else {
					ls->ret = gaamscomm_reader (ccc, CMD_READER_CLEAR);
				}
				if (ls->ret >= 0) {
					if (ls->ret & REP_MASK_CARD) {
						continue;
					}
				}
			}

			if (info.type_node[ccc] == TYPE_MODRL) {
				if (info.enabled[ccc] == 1) {
					ls->ret = gaamscomm_modrl (ccc, mask ? mask : CMD_MODRL_READ);
				} else {
					ls->ret = gaamscomm_modrl (ccc, CMD_MODRL_CLEAR);
				}
			}

			if (info.type_node[ccc] == TYPE_MID) {
				if (info.enabled[ccc] == 1) {
					ls->ret = gaamscomm_mid (ccc, mask ? mask : CMD_MID_READ);
				} else {
					ls->ret = gaamscomm_mid (ccc, CMD_MID_CLEAR);
				}
			}

			if (info.type_node[ccc] == TYPE_SYSTEM) {
				ls->ret = 0;
			}

			if (ls->ret < 0) {
				if (info.enabled[ccc] == 1) {
					p_printf ("Node %02X is no longer connected", info.id_node[ccc]);
				}
				info.enabled[ccc] = 0;
			} else {
				if (info.enabled[ccc] == 2) {
					p_printf ("Node %02X is PRESENT from now on", info.id_node[ccc]);
					info.enabled[ccc] = 1;
				}
				if (is_present == 0 && info.enabled[ccc] == 1) {
					is_present = 1;
				}
			}
		}

	}
	return is_present;
}

void _gaamscomm_answer_thread_push_event (const uint8_t ccc, const uint32_t bitmap, const time_t timestamp, const int clear_counter)
{
	GAAMSCOMM_EVENT ev;

	if (info.debug) p_printf ("THREAD: Pushing event for %02X with bitmap %X", info.id_node[ccc], bitmap);
	ev.num_card = 0;
	ev.ccc = ccc;
	ev.bitmap = bitmap;
	ev.timestamp = timestamp;
	ev.clear_counter = clear_counter;
	queue_push (&info.q_out, &ev);
}

void *_gaamscomm_answer_thread (void *arg)
{
	(void) arg;
	GAAMSCOMM_EVENT ev;
	size_t l = 0;
	FILE *s;
	//struct stat sb;
	uint32_t ret;
	int8_t ret2;
	GAAMSCOMM_CURL crl;
	char buf[PATH_MAX];

#ifndef OS_WIN
	int i, r;
	//uint32_t to_open;
	//int to_port;
	int32_t ret3;
#endif

	if (info.use_what == USE_HTTP) {
		if (http_open (&crl) != 0) {
			p_printf ("ERROR: Failed to initialise cURL, switching to dry run");
			info.use_what = USE_ERROR;
		}
	}

#ifdef OS_WIN
	if (info.use_what == USE_SYSTEM) {
		p_printf ("ERROR: System calls are not supported under WIN32, switching to dry run");
		info.use_what = USE_ERROR;
	}
	if (info.use_what == USE_PIPE) {
		p_printf ("ERROR: Named pipes are not supported under WIN32, switching to dry run");
		info.use_what = USE_ERROR;
	}
#endif

	while (!progend) {
		if (sem_wait (&gaamscomm_answer_semaphore) != 0) break;
		if (progend) break;

		if (queue_front (&info.q_in, &ev) == 0) continue;

		if (abs (ev.timestamp - time (0)) >= MAX_TIME_ELAPSED) {
			p_printf ("THREAD: Received event from %02X, but it's TOO OLD!", info.id_node[ev.ccc]);
			queue_destroy_event (&ev);
			continue;
		}
		if (info.debug) p_printf ("THREAD: Received event from %02X with %d cards", info.id_node[ev.ccc], ev.num_card);
		ret = 0;

		// ***** USE_ERROR *****
		if (info.use_what == USE_ERROR) {
			if (info.type_node[ev.ccc] == TYPE_READER) {
				ret = CMD_READER_RED;
				_gaamscomm_answer_thread_push_event (ev.ccc, ret, ev.timestamp, ev.clear_counter);
			}
		}

		// ***** USE_FILE *****
		if (info.use_what == USE_FILE) {
			if (info.type_node[ev.ccc] == TYPE_READER) {
				buf[0] = '\0';
				for (l = 0; l < ev.num_card; l++) {
					if (l > 0) strcat (buf, " ");
					strcat (buf, ev.card[l]);
				}
				sprintf (buf + strlen (buf), "\n%02X\n", info.id_node[ev.ccc]);

				p_printf ("To file '%s'", info.url);

				pthread_mutex_lock (&gaamscomm_mutex);
				s = fopen (info.url, "w");
				if (s) {
					if (fwrite (buf, 1, strlen (buf), s) != strlen (buf)) {
						ret = CMD_READER_RED;
					} else if (ferror (s)) {
						ret = CMD_READER_RED;
					} else {
						ret = CMD_READER_GREEN;
					}
					fclose (s);
				} else {
					ret = CMD_READER_RED;
				}
				pthread_mutex_unlock (&gaamscomm_mutex);
				_gaamscomm_answer_thread_push_event (ev.ccc, ret, ev.timestamp, ev.clear_counter);
			}
		}

		// ***** USE_PIPE *****
		if (info.use_what == USE_PIPE) {
			ret2 = pipe_send (&ev);
			p_printf ("PIPE request ended with %d", ret2);
			if (ret2 != 0) {
				if (info.type_node[ev.ccc] == TYPE_READER) {
					ret = CMD_READER_RED;
					_gaamscomm_answer_thread_push_event (ev.ccc, ret, ev.timestamp, ev.clear_counter);
				}
			}
		}

		// ***** USE_HTTP *****
		if (info.use_what == USE_HTTP) {
			ret2 = http_send (&crl, &ev);
			p_printf ("HTTP request ended with %d", ret2);
			if (ret2 != 0) {
				if (info.type_node[ev.ccc] == TYPE_READER) {
					if (ret2 == 1) {
						ret = CMD_READER_GREEN;
					} else {
						ret = CMD_READER_RED;
					}
					_gaamscomm_answer_thread_push_event (ev.ccc, ret, ev.timestamp, ev.clear_counter);
				}
			}
		}

		// ***** USE_TCP *****
		if (info.use_what == USE_TCPCLIENT) {
			ret2 = tcpclient_send (&crl, &ev);
			p_printf ("TCP request ended with %d", ret2);
			if (ret2 != 0) {
				if (info.type_node[ev.ccc] == TYPE_READER) {
					if (ret2 == 1) {
						ret = CMD_READER_GREEN;
					} else {
						ret = CMD_READER_RED;
					}
					_gaamscomm_answer_thread_push_event (ev.ccc, ret, ev.timestamp, ev.clear_counter);
				}
			}
		}

		// ***** USE_TCP *****
		if (info.use_what == USE_TCPSERV) {
			ret2 = tcpserver_send (&crl, &ev);
			p_printf ("TCP request ended with %d", ret2);
			if (ret2 != 0) {
				if (info.type_node[ev.ccc] == TYPE_READER) {
					if (ret2 == 1) {
						ret = CMD_READER_GREEN;
					} else {
						ret = CMD_READER_RED;
					}
					_gaamscomm_answer_thread_push_event (ev.ccc, ret, ev.timestamp, ev.clear_counter);
				}
			}
		}

#ifndef OS_WIN
		// ***** USE_SYSTEM *****
		if (info.use_what == USE_SYSTEM) {
			if (info.type_node[ev.ccc] == TYPE_READER) {
				snprintf (buf, PATH_MAX, "%s %02X", info.url + 7, info.id_node[ev.ccc]);
				for (i = 0; i < (int) ev.num_card; i++) {
					strncat (buf, " ", PATH_MAX);
					strncat (buf, ev.card[i], PATH_MAX);
				}
				buf[PATH_MAX - 1] = '\0';

				p_printf ("SYSTEM: %s", buf);
				s = popen (buf, "r");
				if (s) {

					//to_open=0;
					//to_port=-1;
					while (fgets (buf, PATH_MAX, s) != NULL) {
						l = strlen (buf);
						if (l > 0 && buf[0] != '\n' && buf[0] != '\r') {
							if (info.debug) p_printf ("SYSTEM printout: %s", buf);
							ret3 = _gaamscomm_namedpipe_parse (buf);
							p_printf ("_gaamscomm_namedpipe_parse = %d", ret3);
						}
					}
					r = pclose (s);
					p_printf ("SYSTEM returned %d", r);

					if (r > 0) {
						if (r == 256) {
							ret = CMD_READER_GREEN;
						} else {
							ret = CMD_READER_RED;
						}
						_gaamscomm_answer_thread_push_event (ev.ccc, ret, ev.timestamp, ev.clear_counter);
					}
				} else {
					ret = CMD_READER_RED;
					_gaamscomm_answer_thread_push_event (ev.ccc, ret, ev.timestamp, ev.clear_counter);
				}

			}
		}
#endif

		queue_destroy_event (&ev);
	}

	if (info.use_what == USE_HTTP) {
		http_close (&crl);
	}

	pthread_exit (NULL);
	return NULL;
}

int gaamscomm_answer_thread_start (void)
{
	int i;

	queue_clear (&info.q_in);
	queue_clear (&info.q_out);

	gaamscomm_clear_counter++;

	if (sem_init (&gaamscomm_answer_semaphore, 0, 0) != 0) return -1;

	for (i = 0; i < NUM_ANSWER_THREADS; i++) {
		if (pthread_create (&gaamscomm_answer_thr[i], NULL, _gaamscomm_answer_thread, NULL) != 0) return -2;
	}

	return 0;
}

int gaamscomm_answer_thread_stop (void)
{
	int i;

	for (i = 0; i < NUM_ANSWER_THREADS; i++) {
		sem_post (&gaamscomm_answer_semaphore);
	}

	for (i = 0; i < NUM_ANSWER_THREADS; i++) {
		(void) pthread_join (gaamscomm_answer_thr[i], NULL);
	}

	sem_destroy (&gaamscomm_answer_semaphore);
	return 0;
}

#ifndef OS_WIN

int32_t _gaamscomm_namedpipe_parse (char *buf)
{
	char *tstate = NULL, *tok;
	int num;
	uint8_t type = 0;
	int32_t val[8], id = -1, ccc = -1, i;
	uint32_t bitmap;
	num = 0;
	GAAMSCOMM_EVENT ev;

	tok = strtok_r (buf, " \n\r", &tstate);
	while (tok) {
		if (num == 0) {
			if (!strcasecmp (tok, "reader")) type = TYPE_READER;
			if (!strcasecmp (tok, "modrl")) type = TYPE_MODRL;
			if (!strcasecmp (tok, "mid")) type = TYPE_MID;
			if (type == 0) return -1;
		} else if (num == 1) {
			if (sscanf (tok, "%x", &id) != 1) break;
			if (type == TYPE_READER && (id < READER_START || id >= READER_END)) break;
			if (type == TYPE_MODRL && (id < MODRL_START || id >= MODRL_END)) break;
			if (type == TYPE_MID && (id < MID_START || id >= MID_END)) break;
			for (i = 0; i < info.num_nodes; i++) {
				if (info.enabled[i] && info.type_node[i] == type && info.id_node[i] == id) {
					ccc = i;
					break;
				}
			}
		} else if (num <= 9) {
			if (sscanf (tok, "%d", &val[num - 2]) != 1) break;
		} else break;
		tok = strtok_r (NULL, " \n\r", &tstate);
		num++;
	}
	if (ccc < 0 || id < 0) return -2;

	if (type == TYPE_MODRL) {

		if (num < 3) return -3;
		if (val[0] < 0 || val[0] > 13) return -3;

		if (val[0] > 0) {
			if (num < 4) return -3;
			if (val[1] < 0 || val[1] > 6) return -3;

			if (val[1] == 0) {
				bitmap = 1 << 29;
			} else {
				bitmap = val[1] << 29;
			}

			if (val[0] == 13) {
				for (i = 0; i < 13; i++) {
					if (i == 7) continue;
					if (val[1] != 0) {
						bitmap |= 1 << (i + 16);
					}
					bitmap |= 1 << i;
				}
			} else {
				val[0]--;
				if (val[0] > 7) val[0]++; // aby sme prekonali jednu nulu na zaciatku bytu
				if (val[1] != 0) {
					bitmap |= 1 << (val[0] + 16);
				}
				bitmap |= 1 << (val[0]);
			}
		} else {
			bitmap = CMD_MODRL_STATE;
		}

		queue_new_event (&ev);
		ev.ccc = ccc;
		ev.bitmap = bitmap;
		queue_push (&info.q_out, &ev);

	} else if (type == TYPE_READER) {

		if (num < 3) return -3;
		if (val[0] < 0 || val[0] > 3) return -3;

		queue_new_event (&ev);
		ev.ccc = ccc;
		ev.bitmap = (val[0] & 7);
		queue_push (&info.q_out, &ev);

	} else if (type == TYPE_MID) {

		if (num < 3) return -3;
		if (val[0] != 0 && val[0] != 1) return -3;
		bitmap = CMD_MID_READ | ( (val[0] & 1) << 5);

		if (num == 4) {
			if (val[1] != 0 && val[1] != 1) return -3;
			bitmap |= CMD_MID_SET | ( (val[1] & 1) << 15);
		}

		if (num == 5) {
			if (val[2] < 0 || val[2] > 0x7d) return -3;
			bitmap |= CMD_MID_SET | (val[2] << 8);
		}

		queue_new_event (&ev);
		ev.ccc = ccc;
		ev.bitmap = bitmap;
		queue_push (&info.q_out, &ev);

	}

	return 0;
}

void *_gaamscomm_namedpipe_thread (void *arg)
{
	(void) arg;
	int fd, fd2, num;
	char buf[BUFSIZE + 1];
	char *tstate = NULL, *tok;
	int32_t ret;

	umask (0111);
	if (mkfifo (info.namedpipe, 0666) == -1) {
		if (errno != EEXIST) {
			progend = PROGEND_namedPipe;
			return NULL;
		}
	}

	fd = open (info.namedpipe, O_RDONLY | O_NONBLOCK);
	if (fd < 0) {
		progend = PROGEND_namedPipe;
		return NULL;
	}

	fd2 = open (info.namedpipe, O_WRONLY | O_NONBLOCK);
	if (fd2 < 0) {
		close (fd);
		progend = PROGEND_namedPipe;
		return NULL;
	}

	while (!progend) {
		num = read (fd, buf, BUFSIZE);
		if (num < 0) {
			if (errno == EAGAIN || errno == EWOULDBLOCK) {
				usleep (100000);
				continue;
			} else {
				break;
			}
		}
		if (num > 0) {
			buf[num] = '\0';
			tok = strtok_r (buf, "\n\r", &tstate);
			while (tok) {
				if (strlen (tok) > 0) {
					p_printf ("Named pipe: %s", tok);
					ret = _gaamscomm_namedpipe_parse (tok);
					p_printf ("_gaamscomm_namedpipe_parse = %d", ret);
				}
				tok = strtok_r (NULL, "\n\r", &tstate);
			}
		}
		if (num == 0) usleep (100000);
	}

	close (fd2);
	close (fd);
	unlink (info.namedpipe);
	return NULL;
}

int gaamscomm_auto_loop (void)
{
	GAAMSCOMM_LOOP_STATE ls;
	pthread_t namedpipe_thr;

	p_printf ("++++ %s ver. %s ++++ MAIN LOOP STARTING port '%s' ++++", APP_NAME, APP_VER, info.port);

	pthread_mutex_init (&info.q_in.mutex, NULL);
	pthread_mutex_init (&info.q_out.mutex, NULL);

	queue_new (&info.q_in);
	queue_new (&info.q_out);

	gaamscomm_answer_thread_start();
	gaamscomm_clear (&ls);

	if (info.namedpipe[0] != '\0') {
		if (pthread_create (&namedpipe_thr, NULL, _gaamscomm_namedpipe_thread, NULL) != 0) return -2;
	}

	while (!progend) {
		gaamscomm_one_loop (&ls);
	}

	if (info.namedpipe[0] != '\0') {
		(void) pthread_join (namedpipe_thr, NULL);
	}

	gaamscomm_answer_thread_stop();

	pthread_mutex_destroy (&info.q_in.mutex);
	pthread_mutex_destroy (&info.q_out.mutex);

	p_printf ("++++ MAIN LOOP ENDING (progend = %d, readErrorCount = %d, writeErrorCount = %d, noResponseCount = %d) ++++", progend, info.readErrorCount, info.writeErrorCount, info.noResponseCount);
	return 0;
}

#endif // OS_WIN
