#ifndef _GAAMSCOMM_H
#define _GAAMSCOMM_H

#define APP_NAME "GAAMscomm"
#define APP_VER "0.3.31"

#undef OS_WIN
#undef OS_LINUX
#undef OS_MAC

#if defined (_WIN32) || defined (__MINGW32__) || defined (__CYGWIN__)

	#define OS_WIN

	#ifdef WINVER
		#undef WINVER
	#endif // WINVER
	#define WINVER 0x0501

	#ifdef _WIN32_WINNT
		#undef _WIN32_WINNT
	#endif // _WIN32_WINNT
	#define _WIN32_WINNT 0x0501

	#include <winsock2.h>
	#include <ws2tcpip.h>
	#include <io.h>
	#include <windows.h>

	#define RECHECK_AFTER_LOOPS 25
	#define NUM_ANSWER_THREADS 2

#elif defined (__linux__)

	#define OS_LINUX

	#include <sys/types.h>
	#include <sys/socket.h>
	#include <termios.h>
	#include <sys/ioctl.h>
	#include <netinet/ip.h>
	#include <netdb.h>

	#define RECHECK_AFTER_LOOPS 200
	#define NUM_ANSWER_THREADS 4

#elif defined (__APPLE__) && defined (__MACH__)
	#error Apple is not supported in this version
#else
	#error Unsupported platform
#endif

#define HAVE_STRUCT_TIMESPEC
#include <pthread.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/stat.h>
#include <pthread.h>
#include <semaphore.h>

#include <curl/curl.h>
#include <curl/easy.h>

// Reader
#define READER_START 0x01
#define READER_END 0x7E
// Rele modul 12
#define MODRL_START 0x81
#define MODRL_END 0x8F
// Modul vstupov 16
#define MID_START 0x91
#define MID_END 0x9F
// virtualny systemovy modul
#define SYSTEM_START 0xfe
#define SYSTEM_END 0xff

#define BYTE_STX 0xFE

#define BUFSIZE 256
#define CARDBUF_SIZE 6

#define MAX_NODES 128
#define MAX_EVENT_QUEUE 64

#define MAX_TIME_ELAPSED 5

#define MAXreadErrorCount 32
#define MAXwriteErrorCount 8
#define MAXnoResponseCount 256

#define PROGEND_QUIT 1
#define PROGEND_readErrorCount 2
#define PROGEND_writeErrorCount 3
#define PROGEND_noResponseCount 4
#define PROGEND_namedPipe 5

#define TYPE_SYSTEM 0
#define TYPE_READER 1
#define TYPE_MODRL 2
#define TYPE_MID 3

#define USE_ERROR 0
#define USE_FILE 1
#define USE_HTTP 2
#define USE_SYSTEM 3
#define USE_PIPE 4
#define USE_TCPCLIENT 5
#define USE_TCPSERV 6

#define SYSTEM_EVENT_CLEAR 1

typedef enum {
	CARD_OLD,
	CARD_MIFARE,
	_CARD_MAX
} CARD_TYPE;

typedef enum {
	OUTPUT_GAAMS,
	OUTPUT_STRIPPED,
	OUTPUT_REVERSED,
	_OUTPUT_MAX
} CARD_OUTPUT;

typedef struct {
	uint8_t ccc;
	int16_t ret;
	uint32_t look_counter;
} GAAMSCOMM_LOOP_STATE;

typedef struct {
	time_t timestamp;
	uint8_t ccc;
	char *card[CARDBUF_SIZE];
	size_t num_card;
	uint32_t bitmap; // to be OR-ed to the output
	int clear_counter;
} GAAMSCOMM_EVENT;

typedef struct {
	volatile int16_t qbegin, qend;
	GAAMSCOMM_EVENT event[MAX_EVENT_QUEUE];
	pthread_mutex_t mutex;
} GAAMSCOMM_QUEUE;

struct MemoryStruct {
	char *memory;
	size_t size;
	size_t maxsize;
	size_t offset;
};

typedef struct {
	struct MemoryStruct chunk;
	CURL *curl_handle;
	struct curl_slist *curl_slist;

#ifdef OS_WIN
	SOCKET sockfd;
#else
	int sockfd;
#endif
} GAAMSCOMM_CURL;

typedef struct {
	char logpath[PATH_MAX];
	char port[PATH_MAX];
	char entry_id[BUFSIZE];
	char namedpipe[PATH_MAX];

	char *url;
	char *url_proto, *url_host, *url_port, *url_path;

	uint8_t enabled[MAX_NODES];

	uint8_t id_node[MAX_NODES];
	uint8_t type_node[MAX_NODES];
	uint8_t *buf_node[MAX_NODES];
	int num_nodes;

	int8_t use_what;
	char debug, localecho;
	int readErrorCount;
	int writeErrorCount;
	int noResponseCount;

	GAAMSCOMM_QUEUE q_in;
	GAAMSCOMM_QUEUE q_out;

	CARD_OUTPUT card_output;

	struct sockaddr_in serveraddr;

#ifdef OS_WIN
	SOCKET serversockfd;
#else
	int serversockfd;
#endif
} INFO;

/* main.c */
void main_exit (const char *fmt, ...);

/* gaamscomm.c */
void p_printf (const char *fmt, ...);

void gaamscomm_destination_begin (void);
void gaamscomm_destination_end (void);

int8_t gaamscomm_prepare_output (const GAAMSCOMM_EVENT *ev, char *buf, const size_t sz);
int8_t gaamscomm_prepare_tcp_output (const GAAMSCOMM_EVENT *ev, char *buf, const size_t sz);
void gaamscomm_format_card_id (uint8_t *outbuf, CARD_TYPE type, const uint8_t *inbuf, const size_t len);
int gaamscomm_clear (GAAMSCOMM_LOOP_STATE *ls);
int gaamscomm_one_loop (GAAMSCOMM_LOOP_STATE *ls);
int gaamscomm_answer_thread_start (void);
int gaamscomm_answer_thread_stop (void);
int gaamscomm_auto_loop (void);

#ifndef OS_WIN
int32_t _gaamscomm_namedpipe_parse (char *buf);
#endif

/* port.c */
int rs485_open (const char *fname, const int cbaud, const int databits, const int parity);
void rs485_close (void);
ssize_t rs485_send (const ssize_t len, const uint8_t *buf);
ssize_t rs485_recv (uint8_t *buf);
void rs485_debug (const char *text, const ssize_t len, const uint8_t *buf);

/* queue.c */
void queue_new_event (GAAMSCOMM_EVENT *event);
void queue_destroy_event (GAAMSCOMM_EVENT *event);
void queue_new (GAAMSCOMM_QUEUE *q);
int queue_push (GAAMSCOMM_QUEUE *q, GAAMSCOMM_EVENT *ev);
int queue_front (GAAMSCOMM_QUEUE *q, GAAMSCOMM_EVENT *ev);
void queue_clear (GAAMSCOMM_QUEUE *q);

/* http.c */
int8_t http_send (GAAMSCOMM_CURL *c, const GAAMSCOMM_EVENT *ev);
void http_close (GAAMSCOMM_CURL *c);
int8_t http_open (GAAMSCOMM_CURL *c);
int8_t http_global_open (const int use_ssl);
void http_global_close (void);

/* tcpclient.c */
int8_t tcpclient_send (GAAMSCOMM_CURL *c, const GAAMSCOMM_EVENT *ev);
int8_t tcpclient_global_open (void);
void tcpclient_global_close (void);

/* tcpserver.c */
int8_t tcpserver_send (GAAMSCOMM_CURL *c, const GAAMSCOMM_EVENT *ev);
int8_t tcpserver_global_open (void);
void tcpserver_global_close (void);

/* pipe.c */
int8_t pipe_send (const GAAMSCOMM_EVENT *ev);

#endif
