#include "gaamscomm.h"

extern INFO info;

static volatile int _http_use_ssl = 0;

static int _http_SockoptCallback (void *clientp, curl_socket_t curlfd, curlsocktype purpose)
{
	(void) clientp;
	(void) purpose;

#ifdef OS_WIN
	(void) curlfd;
#else
	int tos_local = IPTOS_LOWDELAY;
	int priority = 6;

	if (setsockopt(curlfd, IPPROTO_IP, IP_TOS,  &tos_local, sizeof(tos_local)) != 0) {
		p_printf ("Could not set socket TOS: %s", strerror (errno));
		return 1;
	}

	if (setsockopt(curlfd, SOL_SOCKET, SO_PRIORITY,  &priority, sizeof(priority)) != 0) {
		p_printf ("Could not set socket priority: %s", strerror (errno));
	}
#endif

	return 0;
}

static size_t _http_WriteMemoryCallback (void *ptr, size_t size, size_t nmemb, void *data)
{
	struct MemoryStruct *mem = (struct MemoryStruct *) data;
	size_t realsize = size * nmemb;
	size_t newsize = mem->size + realsize;
	if (newsize >= mem->maxsize - 1) {
		return 0;
	}

	memcpy (& (mem->memory[mem->size]), ptr, realsize);
	mem->size += realsize;
	mem->memory[mem->size] = 0;

	return realsize;
}

int8_t http_send (GAAMSCOMM_CURL *c, const GAAMSCOMM_EVENT *ev)
{
	int ret;
	char buf[4096];

	if ( (ret = gaamscomm_prepare_output (ev, buf, sizeof (buf))) != 0) {
		return ret;
	}
	p_printf ("HTTP POST: %s", buf);
	curl_easy_setopt (c->curl_handle, CURLOPT_POSTFIELDS, buf);

	c->chunk.size = 0;
	if (curl_easy_perform (c->curl_handle) != 0) {
		return -1;
	}
	if (c->chunk.size > 0) {
		c->chunk.memory[c->chunk.size] = '\0';
		p_printf ("HTTP response: %s", c->chunk.memory);
		if (sscanf (c->chunk.memory, "%d", &ret) != 1) {
			return -3;
		}
		if (ret < 0 || ret > 127) {
			return -4;
		}
		return ret;
	}
	return -2;
}

void http_close (GAAMSCOMM_CURL *c)
{
	if (c->curl_handle) {
		curl_easy_setopt (c->curl_handle, CURLOPT_WRITEDATA, NULL);
		curl_easy_cleanup (c->curl_handle);
	}
	if (c->chunk.memory) free (c->chunk.memory);
	curl_slist_free_all (c->curl_slist);
}

int8_t http_open (GAAMSCOMM_CURL *c)
{
	c->chunk.memory = NULL;
	c->curl_handle = NULL;
	c->curl_slist = NULL;

	c->chunk.maxsize = 4096;
	c->chunk.memory = (char *) malloc (c->chunk.maxsize + 5);
	if (!c->chunk.memory) {
		return -2;
	}

	c->curl_handle = curl_easy_init();
	if (!c->curl_handle) {
		return -3;
	}

	/* Fix for lighttpd. Alternative: server.reject-expect-100-with-417 = "disable" */
	c->curl_slist = curl_slist_append (c->curl_slist, "Expect:");

	curl_easy_setopt (c->curl_handle, CURLOPT_TIMEOUT, 4L);
	curl_easy_setopt (c->curl_handle, CURLOPT_CONNECTTIMEOUT, 2L);

	curl_easy_setopt (c->curl_handle, CURLOPT_NOPROGRESS, 1L);
	curl_easy_setopt (c->curl_handle, CURLOPT_POST, 1L);

	curl_easy_setopt (c->curl_handle, CURLOPT_HTTPHEADER, c->curl_slist);
	curl_easy_setopt (c->curl_handle, CURLOPT_WRITEFUNCTION, _http_WriteMemoryCallback);
	curl_easy_setopt (c->curl_handle, CURLOPT_SOCKOPTFUNCTION, _http_SockoptCallback);

	curl_easy_setopt (c->curl_handle, CURLOPT_WRITEDATA, (void *) &c->chunk);
	curl_easy_setopt (c->curl_handle, CURLOPT_USERAGENT, "libcurl-agent/1.0");
	curl_easy_setopt (c->curl_handle, CURLOPT_URL, info.url);

	if (1 == _http_use_ssl) {
		curl_easy_setopt (c->curl_handle, CURLOPT_SSL_VERIFYPEER, 0L);
	}

	return 0;
}

int8_t http_global_open (const int use_ssl)
{
	_http_use_ssl = use_ssl;

#ifdef OS_WIN
	if (1 == use_ssl) {
		if (curl_global_init (CURL_GLOBAL_WIN32 | CURL_GLOBAL_SSL) != 0) {
			return -1;
		}
	}
	else {
		if (curl_global_init (CURL_GLOBAL_WIN32) != 0) {
			return -1;
		}
	}
#else
	if (1 == use_ssl) {
		if (curl_global_init (CURL_GLOBAL_SSL) != 0) {
			return -1;
		}
	}
	else {
		if (curl_global_init (CURL_GLOBAL_NOTHING) != 0) {
			return -1;
		}
	}
#endif

	return 0;
}

void http_global_close (void)
{
	curl_global_cleanup();
}
