#include "../gaamscomm.h"

extern volatile int progend;
extern INFO info;

void handler (int signal)
{
	p_printf ("++++ Caught %s signal ++++", strsignal (signal));
	progend = 1;
}

void main_exit (const char *fmt, ...)
{
	char buf[4096];
	va_list ap;
	va_start (ap, fmt);
	vsnprintf (buf, sizeof (buf), fmt, ap);
	va_end (ap);

	if (strlen (buf) > 0) {
		p_printf ("%s", buf);
		fprintf (stderr, "FATAL ERROR: %s\n", buf);
	}
	exit (EXIT_FAILURE);
}

int main (int argc, char *argv[])
{
	uint32_t nnode;
	//int fd2;

	memset (&info, 0, sizeof (INFO));
	strcpy (info.logpath, "/tmp/gaamscomm");
	mkdir (info.logpath, 0755);

	if (argc < 4) {
		printf ("%s ver. %s\n", APP_NAME, APP_VER);
		printf ("Usage: %s PORT NODES FNAME_OUT [ENTRY_ID] [PIPE_IN]\n", argv[0]);
		printf ("PORT      - Path to RS485 device (usually '/dev/ttyUSB0'). Use prefix:\n\t'@' to enable debug mode\n\t'+' to enable debug mode and local echo too\n\t'-' to enable local echo only\n");
		printf ("NODES     - Node number in hex, can be saparated by spaces\n");
		printf ("FNAME_OUT - Can be either filename, http://, pipe:/, system:/\n");
		printf ("ENTRY_ID  - Value of POST field entry_id used in HTTP(S)\n");
		printf ("PIPE_IN   - Named pipe used to send events to gaamscomm\n");
		return 1;
	}

	char *nodes = strdup (argv[2]);
	if (nodes == NULL) {
		fprintf (stderr, "Malloc error!\n");
		return 2;
	}
	char *state;
	char *tok = strtok_r (nodes, ", ;", &state);
	info.num_nodes = 0;
	while (tok) {
		if (sscanf (tok, "%x", &nnode) != 1) {
			main_exit ("String '%s' is not valid node number!", tok);
		}

		if (info.num_nodes >= MAX_NODES - 1) {
			main_exit ("Too many nodes declared. Maximum is %d!", MAX_NODES - 1);
		}

		if (nnode >= READER_START && nnode < READER_END) {
			info.type_node[info.num_nodes] = TYPE_READER;
		} else {
			if (nnode >= MODRL_START && nnode < MODRL_END) {
				info.type_node[info.num_nodes] = TYPE_MODRL;
			} else {
				if (nnode >= MID_START && nnode < MID_END) {
					info.type_node[info.num_nodes] = TYPE_MID;
				} else {
					main_exit ("Node '%02X' is out of range!", nnode);
				}
			}
		}

		info.id_node[info.num_nodes] = nnode;
		info.num_nodes++;

		tok = strtok_r (NULL, ", ;", &state);
	}
	free (nodes);

	info.id_node[info.num_nodes] = SYSTEM_START;
	info.type_node[info.num_nodes] = TYPE_SYSTEM;
	info.num_nodes++;

	if (argv[1][0] == '@') {
		// We should enable debug
		info.debug = 1;
		strcpy (info.port, argv[1] + 1);
	} else {
		if (argv[1][0] == '+') {
			// We should enable debug and local echo
			info.debug = 1;
			info.localecho = 1;
			strcpy (info.port, argv[1] + 1);
		} else {
			if (argv[1][0] == '-') {
				// We should enable local echo
				info.localecho = 1;
				strcpy (info.port, argv[1] + 1);
			} else {
				strcpy (info.port, argv[1]);
			}
		}
	}

	info.url = strdup (argv[3]);
	if (info.url == NULL) {
		main_exit ("Malloc failed!");
	}

	if (argc >= 5) {
		strncpy (info.entry_id, argv[4], sizeof (info.entry_id));
		info.entry_id[sizeof (info.entry_id) - 1] = '\0';
	} else {
		info.entry_id[0] = '\0';
	}

	if (argc >= 6) {
		strncpy (info.namedpipe, argv[5], sizeof (info.namedpipe));
		info.namedpipe[sizeof (info.namedpipe) - 1] = '\0';
	} else {
		info.namedpipe[0] = '\0';
	}

	// from this point calling of p_printf is allowed
	if (rs485_open (info.port, B38400, CS8, 2) != 0) {
		main_exit ("Could not open port device '%s'", info.port);
	}

	gaamscomm_destination_begin ();

	signal (SIGINT, handler);
	signal (SIGTERM, handler);
	signal (SIGQUIT, handler);

	gaamscomm_auto_loop();

	rs485_close();

	gaamscomm_destination_end ();

	return EXIT_SUCCESS;
}
