#include "gaamscomm.h"

extern volatile int gaamscomm_clear_counter;

void queue_new_event (GAAMSCOMM_EVENT *event)
{
	event->num_card = 0;
	event->clear_counter = gaamscomm_clear_counter;
	event->ccc = 0;
	event->timestamp = time (0);
	event->bitmap = 0;
}

void queue_destroy_event (GAAMSCOMM_EVENT *event)
{
	int i;
	for (i = 0; i < (int) event->num_card; i++) {
		if (event->card[i] != NULL) {
			free (event->card[i]);
			event->card[i] = NULL;
		}
	}
	event->num_card = 0;
}

void queue_new (GAAMSCOMM_QUEUE *q)
{
//    q->mutex=PTHREAD_MUTEX_INITIALIZER;
	q->qbegin = 0;
	q->qend = 0;
}

int queue_push (GAAMSCOMM_QUEUE *q, GAAMSCOMM_EVENT *ev)
{
	GAAMSCOMM_EVENT *t;
	int i;

	if (ev->clear_counter != gaamscomm_clear_counter) return -2;

	pthread_mutex_lock (&q->mutex);

	if (q->qbegin == ( (q->qend + 1) % MAX_EVENT_QUEUE)) {
		pthread_mutex_unlock (&q->mutex);
		p_printf ("QUEUE: FULL, this is serious!");
		return -1; // queue is full
	}

	t = &q->event[q->qend];
	t->bitmap = ev->bitmap;
	t->ccc = ev->ccc;
	t->num_card = ev->num_card;
	t->timestamp = ev->timestamp;
	t->clear_counter = ev->clear_counter;

	for (i = 0; i < (int) t->num_card; i++) {
		t->card[i] = ev->card[i];
		ev->card[i] = NULL;
	}
	ev->num_card = 0;

	q->qend++;
	if (q->qend >= MAX_EVENT_QUEUE) q->qend = 0;
	pthread_mutex_unlock (&q->mutex);
	//p_printf("QUEUE PUSH: %d %d",q->qbegin,q->qend);
	return 0;
}

int queue_front (GAAMSCOMM_QUEUE *q, GAAMSCOMM_EVENT *ev)
{
	GAAMSCOMM_EVENT *t;
	int i;

	pthread_mutex_lock (&q->mutex);

	if (q->qbegin == q->qend) {
		pthread_mutex_unlock (&q->mutex);
		return 0;
	}

	t = & (q->event[q->qbegin]);

	q->qbegin++;
	if (q->qbegin >= MAX_EVENT_QUEUE) {
		q->qbegin = 0;
	}

	if (t->clear_counter != gaamscomm_clear_counter) {
		pthread_mutex_unlock (&q->mutex);
		p_printf ("QUEUE: wrong counter");
		return 0;
	}

	ev->bitmap = t->bitmap;
	ev->ccc = t->ccc;
	ev->num_card = t->num_card;
	ev->timestamp = t->timestamp;
	ev->clear_counter = t->clear_counter;
	for (i = 0; i < (int) ev->num_card; i++) {
		ev->card[i] = t->card[i];
		t->card[i] = NULL;
	}

	pthread_mutex_unlock (&q->mutex);
	//p_printf("QUEUE FRONT: %d %d",q->qbegin,q->qend);
	return 1;
}
/*
void queue_pop(GAAMSCOMM_QUEUE *q)
{
    pthread_mutex_lock(&q->mutex);
    if(q->qbegin!=q->qend){
        q->qbegin++;
        if(q->qbegin>=MAX_EVENT_QUEUE){
            q->qbegin=0;
        }
    }
    pthread_mutex_unlock(&q->mutex);
}
*/
void queue_clear (GAAMSCOMM_QUEUE *q)
{
	pthread_mutex_lock (&q->mutex);
	while (q->qbegin != q->qend) {
		queue_destroy_event (&q->event[q->qbegin]);
		q->qbegin++;
		if (q->qbegin >= MAX_EVENT_QUEUE) {
			q->qbegin = 0;
		}
	}
	pthread_mutex_unlock (&q->mutex);
	queue_new (q);
}
