#include "gaamscomm.h"

extern INFO info;
extern volatile int progend;

pthread_mutex_t _tcpserver_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_t _tcpserver_thr;
#ifdef OS_WIN
	SOCKET _tcpserver_nfds = 0;
#else
	int _tcpserver_nfds = 0;
#endif
fd_set _tcpserver_clients;

int8_t tcpserver_send (GAAMSCOMM_CURL *c, const GAAMSCOMM_EVENT *ev)
{
	(void) c;
	const long timeout_ms = 1500;

	int32_t ret;
	char buf[4096];
	struct timeval tv;
	fd_set infd, outfd, errfd;
	ssize_t iolen;

	if ( (ret = gaamscomm_prepare_tcp_output (ev, buf, sizeof (buf))) != 0) {
		return ret;
	}
	const size_t len = strlen (buf);

	//p_printf ("TCPserv: %s", buf);

	pthread_mutex_lock (&_tcpserver_mutex);

	tv.tv_sec = timeout_ms / 1000;
	tv.tv_usec = (timeout_ms % 1000) * 1000;

	FD_ZERO (&infd);
	outfd = _tcpserver_clients;
	errfd = _tcpserver_clients;

	int res = select (_tcpserver_nfds + 1, &infd, &outfd, &errfd, &tv);
	if (res <= 0) {
		pthread_mutex_unlock (&_tcpserver_mutex);
		return -1;
	}

	ret = -2;

#ifdef OS_WIN
	u_int i;
	for (i = 0; i < outfd.fd_count; i++) {
		iolen = send (outfd.fd_array[i], buf, len, 0);
		if (iolen == SOCKET_ERROR) {
			continue;
		}
		if ((size_t)iolen != len) {
			continue;
		}
		ret = 1;
	}

	for (i = 0; i < errfd.fd_count; i++) {
		closesocket (errfd.fd_array[i]);
		FD_CLR (errfd.fd_array[i], &_tcpserver_clients);
		//printf ("deleting client...\n");
	}
#else
	int i;
	for (i = 0; i <= _tcpserver_nfds; i++) {
		if (!FD_ISSET (i, &outfd)) continue;

		iolen = send (i, buf, len, 0);
		if (iolen < 0) {
			continue;
		}
		else if ((size_t)iolen != len) {
			continue;
		}
		ret = 1;
	}

	for (i = 0; i <= _tcpserver_nfds; i++) {
		if (!FD_ISSET (i, &errfd)) continue;
		close (i);
		FD_CLR (i, &_tcpserver_clients);
	}

#endif // OS_WIN
	pthread_mutex_unlock (&_tcpserver_mutex);

	return ret;
}

void *_tcpserver_thread (void *arg)
{
	(void) arg;
	struct sockaddr_in cli_addr;
	socklen_t cli_len;
#ifdef OS_WIN
	SOCKET newsockfd;
#else
	int newsockfd;
#endif // OS_WIN

	while (!progend) {
		cli_len = sizeof (cli_addr);
		newsockfd = accept (info.serversockfd, (struct sockaddr *) &cli_addr, &cli_len);
#ifdef OS_WIN
		if (newsockfd == INVALID_SOCKET) {
			const int err = WSAGetLastError ();
			p_printf ("TCP accept error %d\n", err);
			break;
		}
#else
		if (newsockfd < 0) {
			p_printf ("TCP accept error %s\n", strerror (errno));
			break;
		}
#endif // OS_WIN

		if (progend) {
			break;
		}

		//printf ("adding new client\n");

		pthread_mutex_lock (&_tcpserver_mutex);
		if (newsockfd > _tcpserver_nfds) {
			_tcpserver_nfds = newsockfd;
		}
		FD_SET (newsockfd, &_tcpserver_clients);
		pthread_mutex_unlock (&_tcpserver_mutex);
	}

	return NULL;
}

int8_t tcpserver_global_open (void)
{
	if (info.url_host == NULL || info.url_port == NULL) {
		return -1;
	}

#ifdef OS_WIN
	WSADATA wsaData;
	if (WSAStartup (MAKEWORD(2,2), &wsaData) != 0) {
		return -1;
	}
#endif

	FD_ZERO (&_tcpserver_clients);

    struct addrinfo hints;
    struct addrinfo *result, *rp;

	memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;

	char host[NI_MAXHOST] = "";
	int ret;

	ret = getaddrinfo(info.url_host, info.url_port, &hints, &result);
	if (ret != 0) {
		p_printf ("DNS error: %s", gai_strerror(ret));
		return -5;
	}

	for (rp = result; rp != NULL; rp = rp->ai_next) {
		if (rp->ai_family == AF_INET) {
			memcpy (&(info.serveraddr), (struct sockaddr_in*)rp->ai_addr, sizeof (struct sockaddr_in));
			ret = getnameinfo (rp->ai_addr, rp->ai_addrlen, host, sizeof (host), NULL, 0, NI_NUMERICHOST);
			if (ret != 0) {
				p_printf ("getnameinfo() failed: %s", gai_strerror (ret));
			}
			p_printf ("DNS translation: %s -> %s", info.url_host, host);
			break;
		}
	}

	if (rp == NULL) {
		freeaddrinfo(result);
		return -6;
	}

	freeaddrinfo(result);

	info.serversockfd = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
#ifdef OS_WIN
	if (info.serversockfd == INVALID_SOCKET) {
		return -1;
	}

	if (bind (info.serversockfd, (LPSOCKADDR)&(info.serveraddr), sizeof(info.serveraddr)) == SOCKET_ERROR) {
		return -2;
	}
#else
	if (info.serversockfd < 0) {
		return -1;
	}

	if (bind (info.serversockfd, (struct sockaddr *)&(info.serveraddr), sizeof(info.serveraddr)) < 0) {
		return -2;
	}
#endif // OS_WIN

	if (listen (info.serversockfd, 5) != 0) {
		return -2;
	}

	pthread_attr_t attr;
	pthread_attr_init (&attr);
	pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);

	if (pthread_create (&_tcpserver_thr, &attr, _tcpserver_thread, NULL) != 0) {
		return -3;
	}

	pthread_attr_destroy (&attr);

	return 0;
}

void tcpserver_global_close (void)
{
#ifdef OS_WIN
	closesocket (info.serversockfd);
	u_int i;
	for (i = 0; i < _tcpserver_clients.fd_count; i++) {
		closesocket (_tcpserver_clients.fd_array[i]);
	}
	WSACleanup ();
#else
	close (info.serversockfd);
	int i;
	for (i = 0; i <= _tcpserver_nfds; i++) {
		if (!FD_ISSET (i, &_tcpserver_clients)) continue;
		close (i);
	}
#endif
}
