#include <allegro.h>
#include <winalleg.h>

#undef _WINSOCKAPI_
#include "../gaamscomm.h"

#include <shlobj.h>
#include <tlhelp32.h>

extern INFO info;
extern volatile int progend;

const char g_szClassName[] = "GAAMscommWindowClass";
const char appName[] = "GAAMscomm";
HICON hIcon, hIconH;
NOTIFYICONDATA tnd;
HMENU pmenu;
HWND hwnd;
int port_loop = FALSE;
int port_open = FALSE;
int check_again = 0;

GAAMSCOMM_LOOP_STATE loop_state;

#define TIMER_DIVISION 5

#define WM_TRAYICON ( WM_USER + 1 )
#define ID_MENU_EXIT 9001

void main_exit (const char *fmt, ...)
{
	char buf[4096];
	va_list ap;
	va_start (ap, fmt);
	vsnprintf (buf, sizeof (buf), fmt, ap);
	va_end (ap);

	if (strlen (buf) > 0) {
		MessageBoxA (NULL, buf, appName, MB_OK | MB_ICONERROR);
	}
	allegro_exit();
	exit (EXIT_FAILURE);
}

int real_path_dir (const int type, char *aFolder)
{
	WCHAR t[PATH_MAX];
	if (SHGetFolderPathW (0, type, NULL, 0, t) != S_OK) return -2;
	int l = WideCharToMultiByte (CP_UTF8, 0, t, lstrlenW (t), aFolder, PATH_MAX, NULL, NULL);
	if (l == 0) return -3;
	aFolder[l] = '\0';
	return 0;
}

HWND main_get_hwnd_from_pid (const DWORD pid)
{
	HWND h = GetTopWindow (0);
	while (h) {
		DWORD apid;
		GetWindowThreadProcessId (h, &apid);
		if (apid == pid) {
			return h;
		}
		h = GetNextWindow (h, GW_HWNDNEXT);
	}
	return NULL;
}

int main_kill_process (const DWORD pid)
{
	HANDLE hProcess;
	HWND h = main_get_hwnd_from_pid (pid);

	if (h) {
		PostMessage (h, WM_CLOSE, 0, 0);
		rest (1000);
	}

	hProcess = OpenProcess (PROCESS_ALL_ACCESS, FALSE, pid);
	if (hProcess) {
		hProcess = OpenProcess (PROCESS_ALL_ACCESS, FALSE, pid);
		if (hProcess == NULL) return FALSE;
		if (!TerminateProcess (hProcess, 1)) {
			CloseHandle (hProcess);
			return FALSE;
		}
		CloseHandle (hProcess);
	}

	return TRUE;
}

int main_check_process (const int shouldKill)
{
	char fullname[PATH_MAX];
	int cnt = 0;

	if (GetModuleFileName (NULL, fullname, PATH_MAX) == 0) return -1;
	char * name = get_filename (fullname);
	if (name == fullname) return -2;

	HANDLE hProcessSnapShot = CreateToolhelp32Snapshot (TH32CS_SNAPALL, 0);

	// Initialize the process entry structure.
	PROCESSENTRY32 ProcessEntry;
	memset (&ProcessEntry, 0, sizeof (ProcessEntry));
	ProcessEntry.dwSize = sizeof (ProcessEntry);

	// Get the first process info.
	BOOL Return = FALSE;
	Return = Process32First (hProcessSnapShot, &ProcessEntry);

	DWORD now = GetCurrentProcessId();

	if (!Return) return -3;
	do {
		if (ProcessEntry.th32ProcessID != now) {
			if (!strcasecmp (name, ProcessEntry.szExeFile)) {
				if (shouldKill) {
					if (!main_kill_process (ProcessEntry.th32ProcessID)) {
						return -4;
					}
				}
				cnt++;
			}
		}
	} while (Process32Next (hProcessSnapShot, &ProcessEntry));
	CloseHandle (hProcessSnapShot);

	return cnt;
}

int main_load_vars (void)
{
	char confname[PATH_MAX];

#ifdef LOCALECHO
	info.localecho = 1;
#else
	info.localecho = 0;
#endif

	packfile_password ("");

	if (GetModuleFileName (NULL, confname, PATH_MAX) == 0) {
		main_exit ("Unable to get current executable path!");
		return 2;
	}
	confname[PATH_MAX - 1] = '\0';
	replace_filename (confname, confname, "gaamscomm.cfg", PATH_MAX);
	confname[PATH_MAX - 1] = '\0';
	set_config_file (confname);

	char *nodes = strdup (get_config_string (NULL, "nodes", "1"));
	if (nodes == NULL) {
		main_exit ("Malloc error!");
		return 2;
	}
	char *tok = strtok (nodes, ", ;");
	info.num_nodes = 0;
	unsigned int nnode;
	while (tok) {
		if (info.num_nodes >= MAX_NODES) {
			main_exit ("Too many nodes declared. Maximum is %d!", MAX_NODES);
			return 2;
		}
		if (sscanf (tok, "%x", &nnode) != 1) {
			main_exit ("String '%s' is not valid node number!", tok);
			return 2;
		}
		if (nnode >= READER_START && nnode < READER_END) {
			info.type_node[info.num_nodes] = TYPE_READER;
		} else {
			if (nnode >= MODRL_START && nnode < MODRL_END) {
				info.type_node[info.num_nodes] = TYPE_MODRL;
			} else {
				if (nnode >= MID_START && nnode < MID_END) {
					info.type_node[info.num_nodes] = TYPE_MID;
				} else {
					main_exit ("Node '%02X' is out of range!", nnode);
					return 2;
				}
			}
		}
		info.id_node[info.num_nodes] = nnode;
		info.num_nodes++;
		tok = strtok (NULL, ", ;");
	}
	free (nodes);

	if (real_path_dir (CSIDL_PERSONAL, info.logpath) == 0) {
		append_filename (info.logpath, info.logpath, "gaamscomm", PATH_MAX);
		mkdir (info.logpath);
	} else {
		info.logpath[0] = '\0';
	}

	if (strcmp (get_config_string (NULL, "debug", "no"), "no")) {
		info.debug = 1;
	} else {
		info.debug = 0;
	}

	port_loop = TRUE;
	strncpy (info.port, get_config_string (NULL, "port", ""), PATH_MAX);
	info.port[PATH_MAX - 1] = '\0';
	if (strlen (info.port) > 0) {
		port_loop = FALSE; // disable port looping
	}

	strncpy (info.entry_id, get_config_string (NULL, "entry_id", ""), BUFSIZE);
	info.entry_id[BUFSIZE - 1] = '\0';

	info.url = strdup (get_config_string (NULL, "fname", ""));
	if (info.url == NULL) {
		main_exit ("Malloc failed!");
	}

	if (strlen (info.url) <= 0) {
		main_exit ("Could not find entry 'fname' in the config file!");
	}

	info.namedpipe[0] = '\0';

	p_printf ("+++ %s version %s start +++", APP_NAME, APP_VER);

	gaamscomm_destination_begin ();

	info.card_output = OUTPUT_GAAMS;
	const char *output_str = get_config_string (NULL, "format", "gaams");
	if (strcasecmp (output_str, "gaams") == 0) {
		info.card_output = OUTPUT_GAAMS;
	} else if (strcasecmp (output_str, "stripped") == 0) {
		info.card_output = OUTPUT_STRIPPED;
	} else if (strcasecmp (output_str, "reversed") == 0) {
		info.card_output = OUTPUT_REVERSED;
	} else {
		main_exit ("Unknown format specified in config file!");
		return 2;
	}

	return 0;
}

LRESULT CALLBACK WndProc (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg) {
		case WM_CREATE:
			pmenu = CreatePopupMenu();
			AppendMenu (pmenu, MF_STRING, ID_MENU_EXIT, "E&xit");
			break;
		case WM_ERASEBKGND:
			break;

		case WM_CLOSE:
			DestroyWindow (hwnd);
			break;

		case WM_DESTROY:
			PostQuitMessage (0);
			break;

		case WM_TIMER:
			if (check_again > 0) {
				check_again--;
				break;
			}
			if (!progend) {
				if (!port_open) {
					int port = 1;
					while (1) {
						if (port_loop) sprintf (info.port, "\\\\.\\COM%d", port);
						if (rs485_open (info.port, CBR_38400, 8, 2) == 0) {
							p_printf ("+++ Opened port %s, trying to find devices +++", info.port + 4);
							if (gaamscomm_clear (&loop_state)) {
								p_printf ("+++ Devices found at port %s +++", info.port + 4);
								port_open = TRUE;

								snprintf (tnd.szTip, sizeof (tnd.szTip), "%s (%s)", appName, info.port + 4);
								tnd.szTip[sizeof (tnd.szTip) - 1] = '\0';
								tnd.hIcon = hIconH;
								Shell_NotifyIcon (NIM_MODIFY, &tnd);
								break;
							} else {
								rs485_close();
							}
						}
						port++;
						if (!port_loop || port > 64) {
							check_again = 15;
							break;
						}
					}
				} else {
					gaamscomm_one_loop (&loop_state);
				}
			} else {
				if (progend == PROGEND_QUIT) DestroyWindow (hwnd);
				else {
					if (port_open) {
						p_printf ("+++ Closing port %s (progend = %d, readErrorCount = %d, writeErrorCount = %d, noResponseCount = %d) +++", info.port, progend, info.readErrorCount, info.writeErrorCount, info.noResponseCount);
						rs485_close();
						port_open = FALSE;

						snprintf (tnd.szTip, sizeof (tnd.szTip), "%s (searching for devices)", appName);
						tnd.szTip[sizeof (tnd.szTip) - 1] = '\0';
						tnd.hIcon = hIcon;
						Shell_NotifyIcon (NIM_MODIFY, &tnd);
						progend = 0;
					}
				}
			}
			break;

		case WM_PAINT:
			break;

		case WM_LBUTTONDOWN:
			break;

		case WM_KEYUP:
			break;

		case WM_TRAYICON:
			if (lParam == WM_RBUTTONDOWN) {
				POINT curPoint;
				GetCursorPos (&curPoint);
				SetForegroundWindow (hwnd);
				TrackPopupMenu (pmenu, TPM_BOTTOMALIGN | TPM_LEFTBUTTON, curPoint.x, curPoint.y, 0, hwnd, NULL);
				//PostMessage(hDlg, WM_NULL, 0, 0);
			}
			break;

		case WM_COMMAND:
			switch (LOWORD (wParam)) {
				case ID_MENU_EXIT:
					progend = PROGEND_QUIT;
					PostMessage (hwnd, WM_CLOSE, 0, 0);
					break;
			}
			break;

		default:
			return DefWindowProc (hwnd, msg, wParam, lParam);
	}
	return 0;
}

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	(void) hPrevInstance;
	(void) lpCmdLine;
	(void) nCmdShow;
	WNDCLASSEX wc;
	MSG Msg;

	set_uformat (U_UTF8);
	if (install_allegro (SYSTEM_NONE, &errno, atexit) != 0) return 99;
	if (main_load_vars() != 0) {
		main_exit ("Initialisation failed!");
		return 1;
	}
	if (main_check_process (FALSE) != 0) {
		main_exit ("Another instance is already running!");
		return 2;
	}

	wc.cbSize        = sizeof (WNDCLASSEX);
	wc.style         = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc   = WndProc;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = hInstance;
	wc.hIcon         = LoadIcon (NULL, IDI_APPLICATION);
	wc.hCursor       = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = g_szClassName;
	wc.hIconSm       = LoadIcon (NULL, IDI_APPLICATION);

	if (!RegisterClassEx (&wc)) {
		return 100;
	}
	hwnd = CreateWindowEx (
			WS_EX_TOPMOST,
			g_szClassName,
			appName,
			WS_POPUP,
			0, 0, 200, 200,
			NULL, NULL, hInstance, NULL);

	if (hwnd == NULL) {
		return 101;
	}

	ShowWindow (hwnd, SW_HIDE);

	hIcon = LoadIcon (hInstance, MAKEINTRESOURCE (101));
	hIconH = LoadIcon (hInstance, MAKEINTRESOURCE (102));

	tnd.cbSize = sizeof (NOTIFYICONDATA);
	tnd.hWnd = hwnd;
	tnd.uID = 0;
	tnd.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
	tnd.uCallbackMessage = WM_TRAYICON;

	snprintf (tnd.szTip, sizeof (tnd.szTip), "%s (searching for devices)", appName);
	tnd.szTip[sizeof (tnd.szTip) - 1] = '\0';

	tnd.hIcon = hIcon;
	Shell_NotifyIcon (NIM_ADD, &tnd);

	pthread_mutex_init (&info.q_in.mutex, NULL);
	pthread_mutex_init (&info.q_out.mutex, NULL);

	queue_new (&info.q_in);
	queue_new (&info.q_out);
	gaamscomm_answer_thread_start();

	SetTimer (hwnd, 101, 1000 / TIMER_DIVISION, (TIMERPROC) NULL);

	while (GetMessage (&Msg, NULL, 0, 0) > 0) {
		TranslateMessage (&Msg);
		DispatchMessage (&Msg);
	}

	gaamscomm_answer_thread_stop();

	if (port_open) {
		p_printf ("+++ QUIT: Closing port %s (progend = %d, readErrorCount = %d, writeErrorCount = %d, noResponseCount = %d) +++", info.port + 4, progend, info.readErrorCount, info.writeErrorCount, info.noResponseCount);
		rs485_close();
		port_open = FALSE;
	}

	Shell_NotifyIcon (NIM_DELETE, &tnd);

	gaamscomm_destination_end ();

	pthread_mutex_destroy (&info.q_in.mutex);
	pthread_mutex_destroy (&info.q_out.mutex);

	allegro_exit();
	if (progend != PROGEND_QUIT) return 1;
	return 0;
}
